package PureMessage::DataStore::cfg;
use PureMessage::Config qw( %config );
use PureMessage::Version qw( $PREFIX );
use PureMessage::Resources::Manager;
use PureMessage::Resources::Store::DB;
use Exporter;
use strict;
our @ISA = qw( Exporter);
our @EXPORT_OK = qw();

sub new {
  my ($class,$name,%options) = @_;
  my $store = PureMessage::Resources::Store::DB->new();
  my $rm = PureMessage::Resources::Manager->new(store => $store);
 
  my $instance_id = delete($options{instance_id});

  die "Creation of a cfg datastore requires an instance_id to be passed to the constructor\n" unless $instance_id;
  # access this resource asif it where a kvdata, we only wrap the
  # kvdata with 'datastore' type so we can give the local instance
  # a different expression
  my $resource = $rm->find($name,type => 'kvdata',instance_id => $instance_id );
  
  die "Unable to find resource '$name' with an instance_id of '$instance_id'" unless $resource;

  my %hash = ( name => $name,
               resource => $resource,
             );

  my $self = bless \%hash, $class;
  return $self;
}

sub keys {
  my $self = shift;
  my $data = $self->{resource}->native();
  my $keys = $data->users();
  return () unless $keys;
  return @$keys; 
}

sub delete {
  my $self = shift;
  my $key = shift;
  my $data = $self->{resource}->native();
  $data->delete_user($key);
}

sub get {
  my $self = shift;
  my $key = shift;
  my $data = $self->{resource}->native();
  return PureMessage::Data::Store::cfg::data->new($data,$key);
}

package PureMessage::Data::Store::cfg::data;

sub new {
  my ($class,$native,$key,%options) = @_;
  my %hash = ( native => $native,
               key => $key,
             );
 
  my $self = bless \%hash, $class;
  return $self;
}


sub store {
  my $self = shift;
  my ($key,$value) = @_;
  $self->{native}->uvalue($self->{key},$key,$value);
  return $key;
}


# store data under a sequentially created unique key. The generated keys are 
# guaranteed to be unique, with their order of creation easily determined by a 
# lexical sort. **note** this will clobber any keys already stored that have
# the same name as 'key_prefix'!!
#
#   a good idiom is 
#   $data->store_seq('key_prefix',$_) foreach qw( a b c d e f g h );
#   my $keys = sort grep /^key_prefix/, $data->contents();

sub store_seq {
  my $self = shift;    
  my ($k_prefix,$value) = @_;
  my $contents = $self->{native}->uvalue($self->{key},$k_prefix);
  die "Contents of $k_prefix is not numeric\n" if (defined($contents) and $contents !~ /^\d+$/);
  $contents = 0 unless $contents; # default to zero if there's no value
  $contents++;
  my $seq_key = sprintf("${k_prefix}_%06d",$contents);
  $self->{native}->uvalue($self->{key},$seq_key,$value);
  $self->{native}->uvalue($self->{key},$k_prefix,$contents);
  return $seq_key; 
}

# return the contents that matches the sequence name
sub contents_seq {
  my ($self,$seq_key) = @_;
  return if not defined($seq_key);
  my @contents = sort grep { /^${seq_key}_/ } $self->contents;
  return @contents if wantarray;
  return \@contents;
}

# returns the available "keys" in this data store
sub contents {
  my $self = shift;
  my $data = $self->{native}->user_keys($self->{key});
  return () unless $data;
  my @keys = keys %$data;
  return @keys if wantarray;
  return \@keys;
}

sub delete {
   my $self = shift;
   my $key = shift;
   $self->{native}->udelete($self->{key},$key);
   return $key;
}

sub retrieve {
  my $self = shift;
  my ($key) = @_;
  return $self->{native}->uvalue($self->{key},$key);
}


1;





