package PureMessage::Resources::CSL;
use strict;
use PureMessage::MessageLog qw(split_line decode_value join_line);
use PureMessage::Resources::Config qw( %resources_config );
use PureMessage::Resources::Location;
use PureMessage::Config qw(expand_filename);
use PureMessage::Version qw($PREFIX);
use PureMessage::Resources::DBH qw(&DBH);
use ActiveState::Config qw(:path);
use ActiveState::File::Atomic;
use base qw(Exporter);

our @EXPORT_OK = qw(&LOG_DEBUG &LOG_INFO &LOG_NOTICE &LOG_WARNING &LOG_ERR &LOG_ERROR &LOG_CRIT &LOG_ALERT &LOG_EMERG);

my $LOG_FILENAME = "$PREFIX/var/log/csl_pending_alerts";
my $LOG_FILENAME_FINAL = "$PREFIX/var/log/csl_log";

# log priority constants (borrowed from syslog via Logger.pm)
sub LOG_DEBUG   () { 0 }
sub LOG_INFO    () { 1 }
sub LOG_NOTICE  () { 2 }
sub LOG_WARNING () { 3 }
sub LOG_ERROR   () { 4 }
sub LOG_ERR     () { 4 }     # alias, compatible with syslog
sub LOG_CRIT    () { 5 }
sub LOG_ALERT   () { 6 }
sub LOG_EMERG   () { 7 }

=head1 NAME

PureMessage::Resources::CSL - Central Server Database Based Error Logging

=head1 SYNOPSIS

  use PureMessage::Resources::CSL;
 
  my $csl = PureMessage::Resources::CSL->new();
  my $location  = PureMessage::Resources::Location->Discover();

  $csl->note(7,"pmx",$location,"error string","lang_id",qw( lang args));
  
  $csl->enumerate( sub { my $a = shift; $a->delete } );


=head1 DESCRIPTION

The C<PureMessage::Resources::CSL> package provides a means of logging errors
safely to the database.  In the event that the database is unavailable it will
log errors to C<var/log/csl_log> and flush them to the database when it next 
becomes available.

=over

=item new

Creates a CSL object you can use to log messages to the database.

=item note($priority,$ident,$location,$error_str,$lang_id,@args)

Logs the passed information to the database as a CSL note. $priority
is one of the exported constants below. $ident is the unique identifier
for the service that is logging the message. $location is a valid
L<PureMessage::Resources::Location> object. $error_str is the text
that should be displayed to the user. $lang_id is a L<PureMessage::Text>
identifier that can be used with the passed array C<@args> to format
a language specific error string.


=item enumerate($CODE)

Enumerates over each CSL note in the database, calling the passed code
reference with a C<PureMessage::Resources::CSL::instance> object
that can be used to manipulate that CSL note in the database.


=item LOG_DEBUG
LOG_INFO
LOG_NOTICE
LOG_WARNING
LOG_ERROR
LOG_ERR
LOG_CRIT
LOG_ALERT
LOG_EMERG

Priority constants that can be imported for use when calling CSL.

=back

=head2 PureMessage::Resources::CSL::instance

=over
  
=item new($csl_id)

Returns a CSL instance. With the exception of the lang_args all CSL
values can be accessed directly:

   my $csl = PureMessage::Resources::CSL::instance->new(1);
   print "$csl->{priority} $csl->{error_str}\n";

=item lang_args

Returns an array of arguments that should be used to format the I<lang_id>

=item delete

Remove this CSL from the database.


=back

=head1 SEE ALSO

L<PureMessage::Resources::Location>

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2003, Sophos PLC
All Rights Reserved.

=cut


sub new {
    my ($CLASS,%conf) = @_;
    my $self;
    $self = bless(\%conf,$CLASS);
    return $self;
}

my $SQL_INSERT_CSL = <<EOF;
   INSERT INTO cfg_alert
        (location_id,hostname,ip,ident,priority,alert_time,info,lang_id,lang_args)
        VALUES (?,?,?,?,?,now(),?,?,?)
EOF

sub note {
    my ($self,@data) = @_;
    $self->note_to_file($LOG_FILENAME_FINAL,@data);
    if ($self->note_to_db(@data) ) {
        $self->flush_pending();
    } else {
        $self->note_to_file($LOG_FILENAME,@data);
    }
}

sub note_to_db {
    my ($self,$priority,$ident,$location,$error_str,$lang_id,@args) = @_;
    my $status = undef;
    eval {
        my $dbh = $self->{use_self_managed_dbh} || DBH();
        die "Invalid DB Handle\n" unless $dbh;
        my @n_args;
	foreach my $arg (@args) {
	    $arg =~ s/\'/\\\'/g;
	    $arg =~ s/\"/\\\"/g;
	    push(@n_args, $arg);
	}
        my $nargs = join ',' , map { "\'$_\'" } @n_args;
        $status = $dbh->do($SQL_INSERT_CSL,undef,$location->location_name || 'unknown',$location->hostname,
                                                   $location->ip,$ident,$priority,
                                                   $error_str,$lang_id,"{$nargs}");
    };
    $self->{db_error} = $@ unless $status;
    return $status;
}

my $SQL_ENUMERATE = <<EOF;
    SELECT alert_id as csl_id
        FROM cfg_alert
        ORDER BY alert_id DESC
EOF

sub query_match {
    my $csl = shift;
    my %query = @_;
    my $match = 1;
    foreach my $key (keys %query) {
        if($key =~ m/^\!(.*)$/ or $key =~ m/^([^!]+)!$/) {
            $match = 0 if exists($csl->{$1}) and $csl->{$1} eq $query{$key};
        } else {
            $match = 0 unless exists($csl->{$key}) and $csl->{$key} eq $query{$key};
        }
    }
    return $match;
}

# query format is field => value or !field => value
sub enumerate {
    my ($self,$CODE,@query) = @_;
    return unless ref($CODE) eq 'CODE';
    my $dbh = DBH();
    my $sth = $dbh->prepare($SQL_ENUMERATE);
    $sth->execute();
    while(my $csl = $sth->fetchrow_hashref()) {
        my $csl = PureMessage::Resources::CSL::instance->new($csl->{csl_id});    
        &$CODE($csl) if query_match($csl,@query);    
    }
}

my $SQL_COUNT = <<EOF;
    SELECT count(alert_id)
        FROM cfg_alert
EOF

my $SQL_COUNT_FILTERED = <<EOF;
    SELECT count(alert_id)
        FROM cfg_alert
        WHERE ident like ?
EOF

sub note_count {
    my ($self,$filter) = @_;
    my $dbh = DBH();
    my $sth;
    if ($filter) {
      $sth = $dbh->prepare($SQL_COUNT_FILTERED);
      $sth->execute($filter);
    } else {
      $sth = $dbh->prepare($SQL_COUNT);
      $sth->execute();
    }
    my $r = $sth->fetchrow_hashref();
    return $r->{count}; 
}

my $SQL_ENUMERATE_IDENT = <<EOF;
    SELECT alert_id as csl_id
        FROM cfg_alert
        WHERE ident like ?
        ORDER BY alert_id DESC
EOF

sub enumerate_ident {
    my ($self,$CODE,$ident) = @_;
    return unless ref($CODE) eq 'CODE';
    my $dbh = DBH();
    my $sth = $dbh->prepare($SQL_ENUMERATE_IDENT);
    $sth->execute($ident);
    while(my $csl = $sth->fetchrow_hashref()) {
        my $csl = PureMessage::Resources::CSL::instance->new($csl->{csl_id});    
        &$CODE($csl);
    }
}


sub note_to_file {
    my ($self,$log,$priority,$ident,$location,$error_str,$lang_id,@args) = @_;
    my @cooked_args = map { (args => $_) } @args;
    open(FV,">>$log") || die "Can't open log file $log";
    print FV join_line( time, hostname => $location->hostname,
                              ip => $location->ip,
                              priority => $priority, 
                              ident => $ident,
                              location => $location->location_name || 'unknown',
                              error_str => $error_str,
                              lang_id => $lang_id,
                              @cooked_args,
                       ) . "\n";
    close(FV) || die "Error closing file $log";
}

sub enumerate_log {
    my ($self,$CODE) = @_;
    my $count = '0ButTrue';
    return unless ref($CODE) eq 'CODE';
    return $count unless -e $LOG_FILENAME_FINAL and -f _ and -s _; #exists and has content
    my $at = ActiveState::File::Atomic->new($LOG_FILENAME_FINAL);
    my %lines;
    my $i = 0;
    $lines{$i++} = $_ while $_ = $at->readline();
    foreach my $line_number (sort { $a <=> $b } keys %lines) {
        my ($ts,@kv) = split_line($lines{$line_number});
        my %line_data = (args => ());
        while(my $a = shift(@kv)) {
            if($a eq 'args') {
               push(@{$line_data{args}},shift(@kv));
            } else {
               $line_data{$a} = shift(@kv);
            }
        }
        # package the location up into a resource again
        my $location= PureMessage::Resources::Location->new(ip => $line_data{ip}, hostname => $line_data{hostname}, location_id => $line_data{location_id});       
        &$CODE(PureMessage::Resources::CSL::instance->new(%line_data,location => $location));    
        $count++;
    }
    $at->close();
    return $count;

}


sub enumerate_pending {
    my ($self,$CODE) = @_;
    my $count = '0ButTrue';
    return unless ref($CODE) eq 'CODE';
    return $count unless -e $LOG_FILENAME and -f _ and -s _; #exists and has content
    my $at = ActiveState::File::Atomic->new($LOG_FILENAME);
    my %lines;
    my $i = 0;
    $lines{$i++} = $_ while $_ = $at->readline();
    foreach my $line_number (sort { $a <=> $b } keys %lines) {
        my ($ts,@kv) = split_line($lines{$line_number});
        my %line_data = (args => ());
        while(my $a = shift(@kv)) {
            if($a eq 'args') {
               push(@{$line_data{args}},shift(@kv));
            } else {
               $line_data{$a} = shift(@kv);
            }
        }
        # package the location up into a resource again
        my $location= PureMessage::Resources::Location->new(ip => $line_data{ip}, hostname => $line_data{hostname}, location_id => $line_data{location_id});       
        &$CODE(PureMessage::Resources::CSL::instance->new(%line_data,location => $location));    
        $count++;
    }
    $at->close();
    return $count;

}


sub flush_pending {
    my $self = shift;
    return unless -e $LOG_FILENAME and -f _ and -s _; #exists and has content
    my $at = ActiveState::File::Atomic->new($LOG_FILENAME, writable => 1);
    my %lines;
    my $i = 0;
    my $count_flushed = '0ButTrue';
    $lines{$i++} = $_ while $_ = $at->readline();
    $self->{use_self_managed_dbh} = PureMessage::Resources::DBH->DBH_SELF_MANAGED();
    foreach my $line_number (sort { $a <=> $b } keys %lines) {
       my %line_data = parse_log_line($lines{$line_number});
       # package the location up into a resource again
        my $location= PureMessage::Resources::Location->new(ip => $line_data{ip}, hostname => $line_data{hostname}, location_id => $line_data{location_id});       
        eval {
            if($self->note_to_db($line_data{priority},$line_data{ident},$location,$line_data{error_str},$line_data{lang_id},@{$line_data{args}})) { # timestamp????
                delete($lines{$line_number});
                $count_flushed++;
            }
        };
        if($@) {
          delete $self->{use_self_managed_dbh};
          die $@ if $@;
        }

    }
    $at->commit_string(join '', map { $lines{$_} } sort keys %lines) if $count_flushed;
    $at->close() unless $count_flushed;
    delete $self->{use_self_managed_dbh};
    return $count_flushed;
}
sub parse_log_line {
		my $data = shift;
		my ($ts,@kv) = split_line($data);
		my %line_data = (args => ());
		while(my $a = shift(@kv)) {
				if($a eq 'args') {
					 push(@{$line_data{args}},shift(@kv));
				} else {
					 $line_data{$a} = shift(@kv);
				}
		}
		return %line_data;
}

1;

package PureMessage::Resources::CSL::instance;
use PureMessage::Resources::DBH qw(DBH);

sub new {
    my ($CLASS,@conf) = @_;
    my $self;
    my %conf;
    # if we are only given a single arg it is the csl_id otherwise it
    # should be the data we need
    if(@conf == 1) {
        $self = bless(\%conf,$CLASS);
        $self->_init(@conf);
    } else {
        %conf = @conf;
        $self = bless(\%conf,$CLASS);
    }
    return $self;
}

my $SQL_CSL_ARGS = <<EOF;
    SELECT lang_args[1], lang_args[2], lang_args[3], lang_args[4], lang_args[5],
           lang_args[6], lang_args[7], lang_args[8], lang_args[9], lang_args[10]
        FROM cfg_alert
        WHERE alert_id = ?
EOF

my $SQL_CSL = <<EOF;
    SELECT location_id, hostname, ip, ident, priority, alert_time, info as error_str, lang_id
        FROM cfg_alert
        WHERE alert_id = ?
EOF



sub _lang_args {
    my $self = shift;
    return $self->{_lang_args} if exists($self->{_lang_args});
    my $dbh = DBH();
    my $sth = $dbh->prepare($SQL_CSL_ARGS);
    $sth->execute($self->{csl_id});
    my $csl = $sth->fetchrow_arrayref();
    return unless $csl;
    my @args = grep(defined($_),@$csl);
    $self->{_lang_args} = \@args;
    return \@args;
}

sub lang_args {
    my $self = shift;
    if (defined($self->{csl_id})) {
        return $self->_lang_args();
    } else {
        return $self->{lang};
    }
}

sub _init {
    my ($self,$csl_id) = @_;
    my $dbh = DBH();
    my $sth = $dbh->prepare($SQL_CSL);
    $sth->execute($csl_id);
    my $csl = $sth->fetchrow_hashref();
    if ($csl) {
        $self->{csl_id} = $csl_id;
        foreach my $key (keys %$csl) {
            $self->{$key} = $csl->{$key};
        }
        $self->{alert_args} = undef;
        my $location;
	if (exists($self->{location_id}) and defined($self->{location_id})) {
	    $location = PureMessage::Resources::Location->Find($self->{location_id});
	}
        if (!$location) {
            # an unregistered location made this CSL note
            $location= PureMessage::Resources::Location->new(ip => $self->{ip}, hostname => $self->{hostname}, location_id => $self->{location_id});       
        }
        $self->{location} = $location;
        
    } else {
        # something bad happened
        die "Unable to initialize a CSL with csl_id of $csl_id";
    } 
}

my $SQL_CSL_DELETE = <<EOF;
   DELETE FROM cfg_alert 
        WHERE alert_id = ?
EOF

sub delete {
    my $self = shift;
    return unless $self->{csl_id};
    my $dbh = DBH();
    return $dbh->do($SQL_CSL_DELETE,undef,$self->{csl_id});
}

1;
