package PureMessage::Resources::DBH;

use strict;
use PureMessage::Config qw(%config);
use ActiveState::Config qw(:path);
use PureMessage::Database::Config qw(cmpdsn);
use PureMessage::Manager::Lang qw(text);
use DBI;
use base 'Exporter';
our @EXPORT = qw( DBH );
our @EXPORT_OK = qw( register_resources_dsn );

sub register_resources_dsn ($);

=head1 NAME

PureMessage::Resources::DBH - Initialize Database Connection

=head1 SYNOPSIS

  use PureMessage::Resources::DBH qw(DBH);
  
  my $dbh = DBH(); 
 
  $dbh->do('insert into cfg_instance (resource_name,profile_id) values ('a',0);


=head1 DESCRIPTION

The C<PureMessage::Resources::DBH> package provides a method to initialize a
database connection. All resource packages should use this package for retrieval
of a database handle as the error handling from the normal C<DBI->connect())
method is much improved in this package. This package makes available the following
class methods.

=over

=item DBH

Returns a database handle if it is possible to be initialized otherwise 
dies with an error.

=cut

our $DBI_HANDLE = undef;
our $PID_ON_CREATE = undef;
our %IN_CSL = ();

use Carp qw(cluck);
# initialize our connection to the database.

sub DBH_SELF_MANAGED {
    my $dsn = $config{store}->{$config{pmx_resources_store}}->{dsn}
	or die "No dsn configured for store '$config{pmx_resources_store}'\n";
    return DBI->connect($dsn,undef,undef,{ PrintError => 0, RaiseError => 1 });
}

sub DBH {
   
   return $DBI_HANDLE if $DBI_HANDLE and $PID_ON_CREATE == $$ and $DBI_HANDLE->ping;
   return if exists($IN_CSL{$$}) and $IN_CSL{$$}; 
   $DBI_HANDLE = undef;
   my $dbh;
   eval {
       sub err {
          my ($munged_error,$object) = @_;
          $IN_CSL{$$} = 1; # block out recursive calls
          my $raw_err = ref($object) ? $object->errstr() : $munged_error;
          my $final_error = '';
          # here is where we could do some magic on the raw handle to get a
          # better error if needed
          my @lines = split("\n",$raw_err);
          foreach my $line (@lines) {
            # discard any newlines and leading whitespace
            chomp($line);
            $line =~ s/^(\s|\t)*//;
            $final_error .= ' ' if $final_error;
            $final_error .= $line;
          }
          $_[0] = "$final_error"; # so it will be raised for us
          # Raise a CSL note
          # pull this in here because otherwise we'll have a cyclic dependancy
          require PureMessage::Resources::CSL;
          require PureMessage::Resources::Location;
          my $csl = PureMessage::Resources::CSL->new();
          my $location  = PureMessage::Resources::Location->new();
          unless($location->location_name()) {
              # they haven't registered this location so we'll have to do
              # a discover to get the location info
              $location  = PureMessage::Resources::Location->Discover();
          }
          $csl->note(PureMessage::Resources::CSL->LOG_ERROR(),"resources",$location,text("dbh_error",$final_error),"dbh_error",$final_error);
          delete($IN_CSL{$$});
          return 0;
          
       }
       my $store = $config{pmx_resources_store};
       my $dsn = $config{store}->{$store}{dsn};
       die "No dsn configured for store '$store'\n" unless $dsn;
       $dbh = DBI->connect_cached($dsn, undef, undef, 
                                  { PrintError => 0, RaiseError => 1, 
                                    HandleError => \&err} );
   };
   # TODO: need to manage error, try and determine why the connect failed
   unless ($dbh) {
        die text('dbh_unable_to_connect',$@) . "\n";
   }

   $PID_ON_CREATE = $$;
   $DBI_HANDLE = $dbh;
   return $dbh
}


# returns 1 if the dsn was added, 0 if it wasn't needed
# should throw an exception on error.

sub register_resources_dsn ($) {
    my $dsn = shift;
    my $as = PureMessage::Database::Config->new(writable => 1);
    my $odsn =
        eval { $as->store_value($config{pmx_resources_store}, "dsn") } || 0;

    my $store_name = (
        $config{pmx_resources_store} eq 'pmx_quarantine' && $odsn &&
        cmpdsn($odsn, $dsn) != 0 # DSNs differ; use different store
    ) ? 'pmx_resources' : $config{pmx_resources_store};

    if($as->update_dsn($dsn, $store_name)) {
        $as->save();
        if($store_name ne $config{pmx_resources_store}) {
            my $as = PureMessage::Config->new(writable => 1);
            path_value($as->root, "pmx_resources_store", $store_name);
            $as->save();
        }
        PureMessage::Config::validate_cache();
        return 1;
    } else {
        $as->unlock;
        return 0;
    }
}


1;



=back

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC

=head1 SEE ALSO
 
 none
  
=cut
