package PureMessage::Resources::KVData::DB;
use strict;
use PureMessage::Resources::Config qw( %resources_config );
use PureMessage::Resources::DBH qw(DBH);
use DBD::Pg;


=head1 NAME

PureMessage::Resources::KVData - Store Key/Value Data in DB 

=head1 SYNOPSIS

  use PureMessage::Resources::Manager;
  use PureMessage::Resources::KVData::DB;

  # create a database store
  my $store = PureMessage::Resources::Store::DB->new();
  my $rmdb = PureMessage::Resources::Manager->new(store => $store);

  $rmdb->Define(type => 'kvdata', name => 'kvtest');
  my $kv = $rmdb->add('kvtest');
  my $data = $kv->native();
  $data->value('bob','silly');
  print $data->value('bob'); #prints silly
    # or
  print $kv->native->value('bob');

=head1 DESCRIPTION

The C<PureMessage::Resources::KVData> provides a key/value store of data.
Objects of this type should not be created manually but instead created
by the C<PureMessage::Resources::Manager>. Available methods are:

=over

=item new($instance_id)

Create a new KVData store. The instance_id can be discoved using the
find method of the C<PureMessage::Resources::Manager>.
object.

   my $kv = PureMessage::Resources::KVData::DB->new(0);
   $kv->native->value('bob','silly');
  

=cut

sub new {
   my ($CLASS,$name,$instance_id) = @_;
   my $self = { name => $name, iid => $instance_id };
   $self = bless($self,$CLASS);
}

sub refresh {
   my $self = shift;
}

# initialize our connection to the database if it's not opened. Otherwise
# return the cached value we have.
sub _dbh {
   my $self = shift;
   return $self->{_dbh} if $self->{_dbh} and $self->{_dbh}->ping;
   $self->{_dbh} = DBH();
   die "Unable to initialize database connection" unless $self->{_dbh};
   return $self->{_dbh};    
}

my $SQL_DELETE_OLD = <<EOF;
  DELETE FROM cfg_kv 
        WHERE instance_id = ?
          AND k_name = ? 
          AND userkey = ?
        
EOF

my $SQL_LOCK_FOR_UPDATE = <<EOF;
  SELECT instance_id FROM cfg_kv
        WHERE instance_id = ?
        AND k_name = ?
        AND userkey = ?
        FOR UPDATE
EOF

my $SQL_UPDATE = <<EOF;
  UPDATE cfg_kv
    SET k_value = ?
    WHERE instance_id = ?
      AND k_name = ?
      AND userkey = ?
EOF

my $SQL_ADD_NEW = <<EOF;
  INSERT INTO cfg_kv (instance_id,k_name,k_value,userkey)
        VALUES (?,?,?::bytea,?)
EOF

my $SQL_GET_VALUE = <<EOF;
  SELECT k_value FROM cfg_kv
        WHERE instance_id = ?
          AND k_name = ?
          AND ? in (userkey,'')
EOF

my $SQL_DELETE_USER = <<EOF;
  DELETE FROM cfg_kv 
        WHERE instance_id = ?
          AND userkey = ?
        
EOF

my $SQL_GET_ALL_USERS = <<EOF;
  SELECT DISTINCT userkey FROM cfg_kv
         WHERE instance_id = ?
EOF

my $SQL_RESET = <<EOF;
  DELETE from cfg_kv
        WHERE instance_id = ?
EOF

my $SQL_GET_ALL_KEYS_PER_USER = <<EOF;
  SELECT k_name, k_value FROM cfg_kv
        WHERE instance_id = ?
          AND userkey = ?
EOF

=item delete($key)

Remove the key/value pair from this kvdata store.

=cut

sub delete {
   my ($self,@input) = @_;
   return $self->udelete($self->_filter(),@input);
}

=item udelete($username,$key)

Removes the key for the passed username

=cut

sub udelete {
   my ($self,$user,@input) = @_;
   my $dbh = $self->_dbh();
   my $status  = $dbh->do($SQL_DELETE_OLD,undef,$self->{iid},$input[0],$user);
   return $status;
}

=item users()

Returns an array ref that contains all defined users.

=cut

sub users {
    my $self = shift;
    my $dbh = $self->_dbh();
    my $sth = $dbh->prepare($SQL_GET_ALL_USERS);
    $sth->execute($self->{iid});
    my @users;
    while( my $data = $sth->fetchrow_hashref()) {
      push(@users,$data->{userkey});
    }
    return \@users;
}

=item reset

Emtpy the kv store of all data

=cut


sub clear_data {
    my $self = shift;
    my $dbh = $self->_dbh();
    return $dbh->do($SQL_RESET,undef,$self->{iid});
}

sub reset {
    my $self = shift;
    return $self->clear_data();
}

=item delete_user($username[,@MoreUserNames])

Remove this user and all their associated data from this store.

=cut

sub delete_user {
   my ($self,@input) = @_;
   my $dbh = $self->_dbh();
   my $status = 1;
   foreach my $user (@input) {
        my $temp_status  = $dbh->do($SQL_DELETE_USER,undef,$self->{iid},$user);
        $status = 0 unless $temp_status; 
   }
   return $status;
}



=item no_user 

Clears the current user so that all internal operations happen on the default
emtpy user.

=cut

sub no_user {
   my $self = shift;
   delete($self->{filter});
}

=item user($username)

Causes all subsequent operations to operate only records that belong to the
passed username.

=cut

sub user {
   my ($self,$filter) = @_;
   $self->{filter} = $filter if defined($filter);
   return $self->{filter};
}

sub _filter {
   my $self = shift;
   if(exists($self->{filter}) and defined($self->{filter})) {
       return $self->{filter};
   } else {
       return '';
   }
}

my $SQL_GET_ALL = <<EOF;
  SELECT k_name, k_value, userkey from cfg_kv
        WHERE instance_id = ?
          AND ? in (userkey,'') 
EOF

my $SQL_GET_ALL_PU = <<EOF;
  SELECT k_name, k_value, userkey from cfg_kv
        WHERE instance_id = ?
EOF

=item hashref()

Returns all the key/value pairs from this kvdata store, taking into account the current
user. Duplicate keys will be expanded to arrays.

=cut

sub hashref {
    my ($self,@input) = @_;
    my $dbh = $self->_dbh();
    my $sth = $dbh->prepare($SQL_GET_ALL);
    $sth->execute($self->{iid},$self->_filter());
    my %h;
    while( my $data = $sth->fetchrow_hashref()) {
      if (exists($h{$data->{k_name}})) {
          my @a = ref($h{$data->{k_name}}) eq 'ARRAY' ? @{$h{$data->{k_name}}} : ($h{$data->{k_name}});
          $h{$data->{k_name}} = [@a,$data->{k_value}];
      } else {
          $h{$data->{k_name}} = $data->{k_value};
      }
    }
    
    return \%h; 
}

=item uhashref()

Returns all the key/value pairs from this kvdata store with a seperate key per
user.

=cut

sub uhashref {
    my ($self,@input) = @_;
    my $dbh = $self->_dbh();
    my $sth = $dbh->prepare($SQL_GET_ALL_PU);
    $sth->execute($self->{iid});
    my %h;
    while( my $data = $sth->fetchrow_hashref()) {
         $h{$data->{userkey}}->{$data->{k_name}} = $data->{k_value};
    }
    return \%h; 
}

=item user_keys($username)

Retruns all the key/value pairs defined for the passed user

=cut

sub user_keys {
    my ($self,$user) = @_;
    $user = '' if !defined($user);
    my $dbh = $self->_dbh();
    my $sth = $dbh->prepare($SQL_GET_ALL_KEYS_PER_USER);
    $sth->execute($self->{iid},$user);
    my %h;
    while( my $data = $sth->fetchrow_hashref()) {
         $h{$data->{k_name}} = $data->{k_value};
    }
    return \%h; 
}

=item value($key[,$value])

If both a value and a key are given sets the associated key to the passed value.
Otherwise returns the value of the passed key.

=cut

sub value {
    my ($self,@input) = @_;
    return $self->uvalue($self->_filter(),@input);
}

=item uvalue($user,$key[,$value]) 

Same as value method except accepts a username as the first paramater and all
operations act on that user.

=cut
sub uvalue {
    my ($self,$user,@input) = @_;
    my $dbh = $self->_dbh();
    if (scalar(@input) == 1) {
        # get the value of the option
        my $sth = $dbh->prepare($SQL_GET_VALUE);
        $sth->execute($self->{iid},$input[0],$user);
        my $data = $sth->fetchrow_hashref();
        return unless $data;
        return $data->{k_value};        
    } elsif (scalar(@input) >= 2) {
        # set the value of the option
        return unless @input % 2 == 0;
        my %data = @input;
        foreach my $key (keys %data) {
            # block out concurrent updates
            my ($data) = $dbh->selectrow_array($SQL_LOCK_FOR_UPDATE,undef,$self->{iid},$key,$user);
            if ($data) {
              # then it already exists in the db update it
              my $sth = $dbh->prepare($SQL_UPDATE);
              $sth->bind_param(1, "$data{$key}", { pg_type => DBD::Pg::PG_BYTEA() });
              $sth->bind_param(2, $self->{iid});
              $sth->bind_param(3, $key);
              $sth->bind_param(4, $user);
              $sth->execute();
            } else {
              # have to insert a new row
              my $sth = $dbh->prepare($SQL_ADD_NEW);
              $sth->bind_param(1, $self->{iid});
              $sth->bind_param(2, $key);
              $sth->bind_param(3, "$data{$key}", { pg_type => DBD::Pg::PG_BYTEA() });
              $sth->bind_param(4, $user);
              $sth->execute();
           }
        }
    }
}

sub autocommit {
    my $self = shift;
    $self->{autcommit} = shift;
}


sub save {
  # this is a no op for the db kv store
}


1;


=back

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC

=head1 SEE ALSO
 
 none
  
=cut
