package PureMessage::Resources::KVData::Local;
use strict;
use PureMessage::Resources::Config qw( %config );
use ActiveState::Config qw(:path);

=head1 NAME

PureMessage::Resources::KVData::Local - Store Key/Value Data in Config 

=head1 SYNOPSIS

  use PureMessage::Resources::Manager;
  use PureMessage::Resources::KVData::Local;

  my $rmdb = PureMessage::Resources::Manager->new();

  $rmdb->Define(type => 'kvdata', name => 'kvtest');
  my $kv = $rmdb->add('kvtest');
  my $data = $kv->native();
  $data->value('bob','silly');
  print $data->value('bob'); #prints silly
    # or
  print $kv->native->value('bob');

=head1 DESCRIPTION

The C<PureMessage::Resources::KVData> provides a key/value store of data.
Objects of this type should not be created manually but instead created
by the C<PureMessage::Resources::Manager>. Available methods are:

=over

=item new($instance_name)

Create a new KVData store. The instance anme can be discoved using the
find method of the C<PureMessage::Resources::Manager>.
object.

   my $kv = PureMessage::Resources::KVData::Local->new('digest-users');
   $kv->native->value('bob','silly');
  

=cut

sub new {
   my ($CLASS,$name,%options) = @_;
   my $self = { name => $name, _hash => undef, attributes => \%options};
   $self = bless($self,$CLASS);
   $self->refresh();
   $self->{autocommit} = exists($options{autocommit}) ? $options{autocommit} : 1;
   $self;
}

sub refresh {
   my $self = shift;
   $self->{as} = undef;
   $self->_init_as(  %{$self->{attributes}} );
}

sub _init_as {
   my $self = shift;
   my %opts = @_;
   $self->{as} = ActiveState::Config->new(strict => 0, allow_includes => 1);
   my $filename = $opts{file} ? $opts{file} : "$config{prefix}/etc/data/resources/$self->{name}.conf";
   unless(-e $filename) {
     open FV,">$filename" or die "Can't create file '$filename' needed by resource '$self->{name}'";
     print FV <<EOF;
# resource sync file for resource $self->{name}
# do not edit this file directory as the contents will be regenerated
# automatically
EOF
    close FV or die "Error accessing file '$filename' for resource '$self->{name}'.  Is the disk full?"
   }
   $self->{as}->parse_file($filename,writable => 1,create => 1);
   $self->{_hash} = undef;
}

=item delete($key)

Remove the key/value pair from this kvdata store.

=cut

sub delete {
   my ($self,@input) = @_;
   return $self->udelete($self->_filter(),@input);
}

=item reset

Emtpy the kv store of all data

=cut

sub clear_data {
   my ($self) = @_;
   my $hash = $self->_as_hash();
   my $as = $self->{as};
   my $root = $as->root();
   foreach my $user (keys %{$hash->{user}}) {
       path_node($root,"user#$user",delete => 1);
   }
   $self->{_data} = undef;
}

sub reset {
   my ($self) = @_;
   $self->clear_data();
   $self->{as}->save();
   $self->refresh();
}

=item udelete($username,$key[,@Morekeys])

Removes the key for the passed username

=cut

sub udelete {
   my ($self,$user,@input) = @_;
   return unless @input;
   my $as = $self->{as};
   my $root = $as->root();
   my $user_key = $self->_user_to_key($user);
   return unless $user_key;
   foreach my $key (@input) {
        path_node($root,"user#$user_key.$key",delete => 1);
   }
   $self->{_data} = undef;
   return 1;
}

=item delete_user($username[,@MoreUserNames])

Remove this user and all their associated data from this store.

=cut

sub delete_user {
   my ($self,@input) = @_;
   return unless @input;
   my $as = $self->{as};
   my $root = $as->root();
   foreach my $user (map { $self->_user_to_key($_) } @input) {
        path_node($root,"user#$user",delete => 1);
   }
   if($self->{autocommit}) {
     $as->save();
     $self->refresh();
     $self->{_data} = undef;
   }
   return 1;
}

sub _as_hash {
    my $self = shift;
    $self->{_data} = $self->{as}->as_hash() unless $self->{_instances};
    return $self->{_data};
}



sub _user_transform {
    my $name = shift;
    $name =~ s/-/--/g;
    $name =~ s/([^a-zA-Z])/_hexit($1)/eg;
    $name;
}

sub _hexit {
  return sprintf("-%xx",ord(shift));
}

sub _user_untransform {
    my $name = shift;
    $name =~ s/--/-/g;
    $name =~ s/-([[\da-f]+)x/chr(hex($1))/eg;
    $name;
}



sub _user_to_key {
   my ($self,$user) = @_;
   return _user_transform($user);
}


=item no_user 

Clears the current user so that all internal operations happen on the default
emtpy user.

=cut

sub no_user {
   my $self = shift;
   delete($self->{filter});
}

=item user($username)

Causes all subsequent operations to operate only records that belong to the
passed username.

=cut

sub user {
   my ($self,$filter) = @_;
   $self->{filter} = $filter if defined($filter);
   return $self->{filter};
}

sub _filter {
   my $self = shift;
   if(exists($self->{filter}) and defined($self->{filter})) {
       return $self->{filter};
   } else {
       return 'ALL_______USERS';
   }
}

=item users()

Returns an array ref that contains all defined users.

=cut

sub users {
    my $self = shift;
    my @values;
    my $h = $self->_as_hash();
    @values = map { my $a = _user_untransform($_); $a eq 'ALL_______USERS' ? '': $a; } keys %{$h->{user}};
    return \@values;
}

=item hashref()

Returns all the key/value pairs from this kvdata store, taking into account the current
user. Duplicate keys will be expanded to arrays.

=cut
sub hashref {
    my ($self,$user) = @_;
    my $hash = $self->_as_hash();
    my %h;
    foreach my $user (keys %{$hash->{user}}) {
      foreach my $k (keys %{$hash->{user}->{$user}}) {
          if (exists($h{$k})) {
              my @a = ref($h{$k}) eq 'ARRAY' ? @{$h{$k}} : ($h{$k});
              $h{$k} = [@a,$hash->{user}->{$user}->{$k}];
          } else {
              $h{$k} = $hash->{user}->{$user}->{$k};
          }
      }
    }
    return \%h;
}

=item uhashref()

Returns all the key/value pairs from this kvdata store with a seperate key per
user.

=cut

sub uhashref {
    my ($self,@input) = @_;
    my $hash = $self->_as_hash();
    my %new;
    foreach my $user (keys %{$hash->{user}}) {
        my $t = _user_untransform($user);
        $t = '' if $t eq 'ALL_______USERS';
        $new{$t} = $hash->{user}->{$user};
    }
    return \%new;
}
=item user_keys($username)

Retruns all the key/value pairs defined for the passed user

=cut

sub user_keys {
    my ($self,$user) = @_;
    $user = $self->_filter() if $user eq '' or !defined($user);
    my $tuser = _user_transform($user);
    my $hash = $self->_as_hash();
    return unless exists($hash->{user}->{$tuser});
    my %new;
    foreach my $key (keys %{$hash->{user}->{$tuser}}) {
        $new{$key} = $hash->{user}->{$tuser}->{$key};
    }
    return \%new;
}


=item value($key[,$value])

If both a value and a key are given sets the associated key to the passed value.
Otherwise returns the value of the passed key.

=cut

sub value {
    my ($self,@input) = @_;
    return $self->uvalue($self->_filter(),@input);
}

=item uvalue($user,$key[,$value]) 

Same as value method except accepts a username as the first paramater and all
operations act on that user.

=cut
sub uvalue {
    my ($self,$user,@input) = @_;
    my $tu = $self->_user_to_key($user);
    if (scalar(@input) == 1) {
        my $hash = $self->_as_hash();
        # get the value of the option
        return unless exists($hash->{user}->{$tu});
        return $hash->{user}->{$tu}->{$input[0]} if exists($hash->{user}->{$tu}->{$input[0]});
        return;        
    } elsif (scalar(@input) >= 2) {
        return unless @input % 2 == 0; # must be k/v
        my %data = @input;
        # set the value of the option
        my $as = $self->{as};
        my $root = $as->root();
        foreach my $key (keys %data) {
            path_value($root,"user#$tu.$key",$data{$key});
        }
        if ($self->{autocommit}) {
            $as->save();
            $self->refresh();
            $self->{_data} = undef;
        }
    }
}

# set the autocommit flag, this is needed because some consumers
# of this package make aand cache this object and may therefore not 
# be able to know they should autocommit or not
sub autocommit {
    my $self = shift;
    $self->{autcommit} = shift;
}

sub save {
   my $self = shift;
   $self->{as}->save();
   $self->{_data} = undef;
   $self->refresh();
}

1;


=back

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC

=head1 SEE ALSO
 
 none
  
=cut



