package PureMessage::Resources::Manager;
use strict;
use PureMessage::Resources::Config qw( %resources_config );
use PureMessage::Resources::Store::DB;
use PureMessage::Resources::Store::Local;
use PureMessage::Resources::DBH qw(DBH);
use PureMessage::Resources::Resource;
use ActiveState::Config qw(:path);
use ActiveState::Config::Simple qw(config_parse_file);
use PureMessage::Resources::Log;

=head1 NAME

PureMessage::ResourcesManager - Manage the creation and access of resources

=head1 SYNOPSIS

  use PureMessage::Resources::Manager;
  
  # create a database store
  my $store = PureMessage::Resources::Store::DB->new();
  my $rmdb = PureMessage::Resources::Manager->new(store => $store);

  # create a default store
  my $rmlocal = PureMessage::Resources::Manager->new();
  
  

=head1 DESCRIPTION

The C<PureMessage::Resources::Manager> provides a general functionality to discover
and manipulate resources.

=over

=item new(store => $store, %options)

Create a new resource manager. If a resource store that is different from the 
Local store is required it should be created prior to this call and passed
in as the store option. eg.

   my $store = PureMessage::Resources::Store::DB->new();
   my $rm = PureMessage::Resources::Manager->new(store => $store);
   

=cut

our $__dbh;

my $SQL_INSERT_RESOURCE_DEFINITION = <<EOF;
     insert into cfg_resource(resource_name,resource_type)
       values(?,?)
EOF

my $SQL_INSERT_RESOURCE_ATTRIBUTES = <<EOF;
     insert into cfg_attribute(resource_name,attribute_name,attribute_value)
       values(?,?,?)
EOF


my @VALID_RESOURCE_TYPES = (
    'list', 'multilist', 'kvdata', 'asdata', 'datastore',
    'map', 'preferences', 'session', 'blob'
);

my $SQL_SEARCH_FOR_RESOURCE = <<EOF;
 SELECT resource_name, resource_type 
        FROM cfg_resource
        WHERE resource_name = ?

EOF

my $SQL_SEARCH_FOR_RESOURCE_ATTRIBUTES = <<EOF;
 SELECT attribute_name,attribute_value 
        FROM cfg_attribute
        WHERE resource_name = ?

EOF

my $SQL_GET_ALL_RESOURCE_DEFS = <<EOF;
  SELECT resource_name, resource_type 
         FROM cfg_resource
EOF

my $SQL_DELETE_RESOURCE_DEF = <<EOF;
  DELETE from cfg_resource 
        WHERE resource_name = ?
EOF

sub new {
    my ($CLASS,%conf) = @_;
    my $store = delete($conf{store});

    # utilize the global setting to override the default store
    # type we create
    if(!$store and $resources_config{default} eq 'db') {
        require PureMessage::Resources::Store::DB;
        $store = PureMessage::Resources::Store::DB->new(%conf);
    }
    unless($store) {
        require PureMessage::Resources::Store::Local;
        $store = PureMessage::Resources::Store::Local->new(%conf);    
    }
    my $self;
    $self->{store} = $store;
    # it is possible to have the store passed in to us, which means
    # that we need to change the passed store to writable if they've
    # passed the writable flag to us but not the store
    $self->{store}->writable($conf{writable}) if exists($conf{writable});
    return bless($self,$CLASS);
}

=item Define(%options)

Define a resource with the passed options. The location the definition
is kept is determined by the value of the C<default> setting of the
I<resources.conf> config file.  Option values are:

  type => resource type, as_config, list, map, mlist, kvdata
  name => resource name
  attributes => \%settings

You can optionally force the location that the resource is stored in by
passing an option of:

  store => local|db
   

=cut

sub Define {
   my ($CLASS,%resource_def) = @_;
   $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
   my $forced_location = delete($resource_def{'store'}) || $resources_config{default};
   if($forced_location ne 'db') {
      # we should store this locally
      return PureMessage::Resources::Manager->DefineLocal(%resource_def);
   } else {
      return PureMessage::Resources::Manager->DefineDB(%resource_def);
   }
   
}

sub DefineLocal {
   my ($CLASS,%resource_def) = @_;
   $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name  
   # if we are handed an AS::Config object use that for writing instead, 
   # also don't save it but modify it instead, also fakeup our Config file hash
   # so we don't have to refresh it
   my $as = delete($resource_def{_as_config});
   my $auto_commit = $as ? 0 : 1; # if yes we save automatically but don't if we get an AS
                                  # object as input
   
   my %attributes = ();
   if (ref($resource_def{attributes} eq 'HASH')) {
     %attributes = %{$resource_def{attributes}};
   }
  
   die "invalid resource name '$resource_def{name}'\n" unless $resource_def{name} =~ /^.+$/;

   # in the future.
   die "invalid resource type '$resource_def{type}'\n" unless grep($_ eq $resource_def{type},@VALID_RESOURCE_TYPES);

   # don't add a resource definition if it's already there
   my $found_data = $CLASS->FindLocal($resource_def{name}); # undef if not found
   return if $found_data;
   
   unless($as) {
     $as = PureMessage::Resources::Config->edit();
   }
   
   my $root = $as->root();
   my $name = $resource_def{name};
   foreach my $key (keys %resource_def) {
       # XXX: TODO: this block forces us to support resources defs with only one nested
       # config block.  This stops us for instance from supporting ldap: list
       # types because they have 3 levels of nesting
       if (ref($resource_def{$key}) eq 'HASH') {
           foreach my $subkey (keys %{$resource_def{$key}}) {
               path_value($root,"resource#$name.$key.$subkey",$resource_def{$key}->{$subkey});
               # update our config cache so it's current
               $resources_config{resource}->{$name}->{$key}->{$subkey} = $resource_def{$key}->{$subkey};
           }
       } else {
           path_value($root,"resource#$name.$key",$resource_def{$key});
           # update our config cache so it's current
           $resources_config{resource}->{$name}->{$key} = $resource_def{$key};
       }
   }
   
   if($auto_commit) {
     $as->save();
     PureMessage::Resources::Config->refresh();
   }

   return \%resource_def;
      
}


sub DefineDB {
   my ($CLASS,%resource_def) = @_;
   $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
   # TODO this breaks if there are no attributes required
   my %attributes = %{$resource_def{attributes}};
  
   # TODO tighten this up before stuffing into the db
   die "invalid resource name '$resource_def{name}'\n" unless $resource_def{name} =~ /^.+$/;

   # TODO make sure the resource type is valid. This should be kept in the database
   # in the future.
   die "invalid resource type '$resource_def{type}'\n" unless grep($_ eq $resource_def{type},@VALID_RESOURCE_TYPES);

   # don't add a resource definition if it's already there
   my $found_data = $CLASS->FindDB($resource_def{name}); # undef if not found
   return if $found_data;

   # TODO try to add this resource record now, be pretty nice to wrap this into a
   # transaction...
   my $dbh = $CLASS->_dbh();
   my $status = $dbh->do($SQL_INSERT_RESOURCE_DEFINITION,undef,$resource_def{name},$resource_def{type});
   die "Unable to insert new resource definition\n" unless $status;
   # try to insert the attributes associated with this resource into the db
   my $attrs;
   my $sth = $dbh->prepare($SQL_INSERT_RESOURCE_ATTRIBUTES);

   foreach my $key (keys %attributes) {

       # TODO this needs to be tighted up too
       die "invalid attribute name '$key' for resource '$resource_def{name}'.\n" unless $key =~ /^.+$/;

       die "error inserting attribute '$key' for resource '$resource_def{name}'\n" unless 
           $sth->execute($resource_def{name},$key,$attributes{$key});
   }
                    
   return \%resource_def;
}

=item DefineFromFile($filename,[%options])

Define all the resources from the passed filename.  This will define the 
resources into the current C<default> store as defined by the 
I<resources.conf> configuration file.  This is primarily useful when first
bootstrapping the system.  You can force initialization into a specific
store by passing an option of C<store => db|local>.

=cut


sub DefineFromFile {
    my ($CLASS,$filename,%options) = @_;
    $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
    # only make an absolute filename if it isn't
    $filename = "$resources_config{prefix}/etc/$filename" unless $filename =~ m|^/|;
    my %conf = %{config_parse_file($filename)}; 
    my $count = 0;
    my $as = undef;
    foreach my $res (sort keys %{$conf{resource}}) {
       next if PureMessage::Resources::Manager->Find($res,%options);
       $as = PureMessage::Resources::Config->edit() unless $as;
       # fixeup the passed resources def since name isn't required in *both*
       # locations in the conf file
       $conf{resource}->{$res}->{name} = $res unless $conf{resource}->{$res}->{name};
       $count++ if PureMessage::Resources::Manager->Define(%{$conf{resource}->{$res}},%options, _as_config => $as );
    }
    if($count && $as) {
      # then commit those changes to disk and refresh our view of that data
      $as->save();
      PureMessage::Resources::Config->refresh();
    }
    return $count;
}

=item initFromFile($filename)

Create an instance of the resources from the passed filename.  This will create
instances into the current C<default> store as defined by the 
I<resources.conf> configuration file.  This is primarily useful when first
bootstrapping the system.

=cut


sub initFromFile {
    my ($self,$filename,%options) = @_;
    my $as = ActiveState::Config->new(strict => 0, allow_includes => 1);
    # only make an absolute filename if it isn't
    $filename = "$resources_config{prefix}/etc/$filename" unless $filename =~ m|^/|;
    $as->parse_file($filename);
    my @added = ();
    my %conf = $as->as_hash;
    my $count = 0;
    my @required = ref($conf{require}) eq 'ARRAY' ? @{$conf{require}} : ($conf{require});
    $self->begin_batch();
    foreach my $res (sort @required) {
       next if $self->find($res,%options);
       if ($self->add($res)) {
          $count++;
          push(@added,$res);
       }
    }
    $self->end_batch();
    my $rconfig = PureMessage::Resources::Config->edit();
    my @imports = path_values($rconfig->root,"import");
    my @exports  = path_values($rconfig->root,"export");

    if( ref($conf{'import'}) eq 'ARRAY') {
        foreach my $k (@{$conf{'import'}}) {
          next if grep($k eq $_,@imports);
          push(@imports,$k);
        }
        path_values($rconfig->root,"import",@imports);
    }

    if( ref($conf{'export'}) eq 'ARRAY') {
        foreach my $k (@{$conf{'export'}}) {
          next if grep($k eq $_,@exports);
          push(@exports,$k);
        }
        path_values($rconfig->root,"export",@exports);
    }
    $rconfig->save();
    return wantarray ? @added : $count;
}




# initialize our connection to the database if it's not opened. Otherwise
# return the cached value we have.
sub _dbh {
   my $CLASS = shift;
   $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
   {
       no strict 'refs';
       return ${"${CLASS}::__dbh"}
           if(${"${CLASS}::__dbh"} and ${"${CLASS}::__dbh"}->ping);

       ${"${CLASS}::__dbh"} = DBH();
       die "Unable to initialize database connection" unless ${"${CLASS}::__dbh"};
       return ${"${CLASS}::__dbh"};
   };
}

=item Available([%options])

Returns a hash or a ref to hash of available resources and their types. To override
the default store pass an option of:
  
  store => local|db

=cut

sub Available {
    my ($CLASS,%options) = @_;
    $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
    my $forced_location = delete($options{'store'}) || $resources_config{default};
    
    if($forced_location ne 'db') {
       # we should store this locally
       return PureMessage::Resources::Manager->AvailableLocal(%options);
    } else {
       return PureMessage::Resources::Manager->AvailableDB(%options);
    }
   

}


sub AvailableLocal {
    my ($CLASS,%options) = @_;
    $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
    my %available;
    foreach my $key (keys %{$resources_config{resource}}) {
        my $data = $resources_config{resource}->{$key};
        if ($options{bytype}) {
            $available{$data->{type}} = [] unless exists($available{$data->{type}});
            push(@{$available{$data->{type}}},$data->{name});
        } else {
            $available{$data->{name}}->{type} = $data->{type};
        }
    }
    return wantarray ? %available : \%available;
}

# this will remove the resource definition we have in the DB
sub DeleteResourceDefinition {
  my ($self,@resources) = @_;
  my $LOG = PureMessage::Resources::Log->new();
  my $dbh = $self->_dbh();
  foreach my $r (@resources) {
    #warn "attempt to delete $r\n"; 
    my $status = $dbh->do($SQL_DELETE_RESOURCE_DEF,undef,$r);
    $LOG->log('INFO',"Removing resource definition for resource '$r' from the DB store."); 
  }
}

sub AvailableDB {
    my ($CLASS,%options) = @_;
    $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
    my $dbh = $CLASS->_dbh();
    my $sth = $dbh->prepare($SQL_GET_ALL_RESOURCE_DEFS);
    $sth->execute();
    my %available;
    while (my $data = $sth->fetchrow_hashref() ) {
        if ($options{bytype}) {
            $available{$data->{resource_type}} = [] unless exists($available{$data->{resource_type}});
            push(@{$available{$data->{resource_type}}},$data->{resource_name});
        } else {
            $available{$data->{resource_name}}->{type} = $data->{resource_type};
        }
    }
    return wantarray ? %available : \%available;
}

=item Find($resource_name,[%options])

Finds and returns a hash ref containing the particulars for the named resource.
Note this is the actual definition of the resource I<not> the instance
of this resource. If the named resource has not been defined, returns undef.

   my $resource_def = PureMessage::Resources::Manager->Find('white-list);

To force the search to a specific store pass an option of:

   store => local | db

=cut

# returns a \%resource_def or undef
sub Find {
    my ($CLASS,$name,%options) = @_;
    $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
    my $forced_location = delete($options{'store'}) 
                          || $resources_config{default} || '';
    if($forced_location ne 'db') {
       # we should store this locally
       return PureMessage::Resources::Manager->FindLocal($name);
    } else {
       return PureMessage::Resources::Manager->FindDB($name);
    }
   

}

sub FindLocal {
    my ($CLASS,$name) = @_;
    $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
    if(exists($resources_config{resource}->{$name}) and $resources_config{resource}) {
       my $temp = $resources_config{resource}->{$name};
       return $temp;
    } else {
       return;
    }
}


# returns a \%resource_def or undef
sub FindDB {
    my ($CLASS,$name) = @_;
    $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
    # attempt a db connection only when it's required.    
    my $dbh = $CLASS->_dbh();
    my $sth = $dbh->prepare($SQL_SEARCH_FOR_RESOURCE);
    $sth->execute($name);
    # there should only be ONE of these so we'll discount the
    # case were we retrieve two of them
    my $data = $sth->fetchrow_hashref();
    return unless $data;
    # start collecting our data for return
    my %resource_def = ( type => $data->{resource_type},
                         name => $data->{resource_name},
                         attributes => {},
                       );
    
    # get the Attributes then!
    $sth = $dbh->prepare($SQL_SEARCH_FOR_RESOURCE_ATTRIBUTES);
    $sth->execute($name);

    while($data = $sth->fetchrow_hashref) {

       my ($name,$value) = ($data->{attribute_name},$data->{attribute_value});

       if( exists( $resource_def{attributes}->{$name} ) ) {

          # then we might need to promote the value to an array 
          my @new_value = ref(  $resource_def{attributes}->{$name} ) eq 'ARRAY' 
                           ? @{ $resource_def{attributes}->{$name} } 
                           : (  $resource_def{attributes}->{$name} );
          push(@new_value,$value);
          $resource_def{attributes}->{$name} = \@new_value;

       } else {
          $resource_def{attributes}->{$name} = $value;
       }

    }
    return \%resource_def;
}

=item find($resource_name)

Find and return the C<PureMessage::Resources::Resource> object for
the named resource.

=cut

sub find {
    my ($self,@args) = @_;
    return $self->{store}->find(@args);
}

sub begin_batch {
    my ($self,@args) = @_;
    return $self->{store}->begin_batch(@args);
}

sub end_batch {
    my ($self,@args) = @_;
    return $self->{store}->end_batch(@args);
}


=item forget($resource_name)

Cause the store to delete any cached references to the given resource
it was maintaining. This may be neccessary if the resource you are
using has local exclusive locks. eg. preferences

=cut


sub forget {
    my ($self,@args) = @_;
    return $self->{store}->forget(@args);
}

=item add($resource_name)

Add an instance of the named resource to the currently configured resource
store.

=cut

sub add {
    my ($self,@args) = @_;
    return $self->{store}->add(@args);
}

=item remove($resource_name)

Remove the instance of the named resource from the currently 
managed store.

=cut

sub remove {
    my ($self,@args) = @_;

    my $LOG = PureMessage::Resources::Log->new();
    $LOG->log('INFO',"Removing resource instance '$_' from " . $self->{store}->type() . " store")
        foreach @args;

    return $self->{store}->remove(@args);
}

=item enumerate($&sub)

For each resource intance contained in the currently managed resource
store call the passed subroutine with a C<PureMessage::Resources::Resource>
object that refers to it. An example usage:

  my $sub = sub {
                    my $resource = shift;
                    print $resource->name . "\n";
                };
                
  $manager->enumerate( $sub ); 


=cut

sub enumerate {
    my ($self,@args) = @_;
    return $self->{store}->enumerate(@args);
}

=item sync(@resource_objects)

Sync the contents of the resources contained in the passed 
C<PureMessage::Resources::Resource> objects.

  my $resource = $remote_manager->find('white-list');
  $local_manager->sync($resource);

=cut

sub sync {
    my ($self,@args) = @_;
    return $self->{store}->sync(@args);
}

=item flush_item_change

Cleanup unneeded or old rows from the chg_item_change table. This
should be called periodically as a required maintenance activity.

=cut


my $SQL_FLUSH_ITEM_CHANGE_PENDING = <<EOF;

SELECT instance_id, min(last_change_id) as last_change_id 
  FROM cfg_location_instance GROUP BY instance_id

EOF


my $SQL_FLUSH_ITEM_CHANGE = <<EOF;

DELETE FROM cfg_item_change 
 WHERE instance_id = ?
   AND change_id < ?
EOF


my $SQL_FLUSH_ITEM_CHANGE_ALL_CHANGE_ID = <<EOF;
DELETE FROM cfg_item_change
 WHERE instance_id = ?
EOF

my $SQL_FLUSH_ITEM_CHANGE_MAX = <<EOF;

DELETE FROM cfg_item_change
  WHERE change_id < 
        ( SELECT CASE 
                WHEN max(change_id) - ? < 0 THEN 0 
                ELSE max(change_id) - ? END AS ccid FROM cfg_item_change
        )

EOF

sub flush_item_change {
   my ($CLASS,%resource_def) = @_;
   $CLASS = ref($CLASS) if ref($CLASS); # if it's an object get the class name   
   
   my $dbh = $CLASS->_dbh();
   my $sth = $dbh->prepare($SQL_FLUSH_ITEM_CHANGE_PENDING);
   $sth->execute();
   while(my $data = $sth->fetchrow_hashref() ) {
       if($data->{last_change_id}) {
         $dbh->do($SQL_FLUSH_ITEM_CHANGE,undef,$data->{instance_id},$data->{last_change_id});
       } else {
         $dbh->do($SQL_FLUSH_ITEM_CHANGE_ALL_CHANGE_ID,undef,$data->{instance_id});
       }
   }
   # harvest changes that are greater than our limit
   # the CSL notes system will raise any errors we have with this sql
   $dbh->do($SQL_FLUSH_ITEM_CHANGE_MAX,undef,$resources_config{max_item_change},$resources_config{max_item_change});
 
}



=item change_id($resource_name)

Will find the change_id for the passed resource_name as it
is recorded in the store.

=cut

sub change_id {
  my ($self,$resource_name) = @_;
  return $self->{store}->change_id($resource_name);
}

=item store_type

Returns the type of store this manager is managing.

=cut

sub store_type {
   my $self = shift;
   return $self->{store}->type();
}



DESTROY {
  my $self = shift;
  $self->{_dbh}->disconnect() if $self->{_dbh};
}

sub _get_unique_id {
  my $self = shift;
  my $dbh = $self->_dbh();
  my $id = $dbh->selectall_arrayref("select nextval('cfg_resource_ident')")->[0][0];
  die "Can't generate a unique id for a resources.  Cannot continue." unless $id;
  return $id;
}

sub CreateResourceTarget {
  my ($CLASS,$type,$list_class,%options) = @_;
  die unless $type eq 'list' or $type eq 'map';
  
  my $LOG = PureMessage::Resources::Log->new();
  
  $list_class  = "custom" unless $list_class;
  my %resource_def;
  
  # make up a name
  my $name_template = $list_class . '_%04d';
  my $ix = $CLASS->_get_unique_id();
  my $resource_name = sprintf($name_template,$ix);
 
  my %defined_resources = PureMessage::Resources::Manager->Available(store => 'db');
  while(exists($defined_resources{$resource_name})) {
    $ix++;
    $resource_name = sprintf($name_template,$ix);
  }
  $resource_def{name} = $resource_name;
  $resource_def{type} = $type;
  $resource_def{store} = 'db';
  my %attr;
  $attr{name} = $options{name} || $resource_name;
  $attr{description} = $options{description} || $resource_name;
  $attr{source} = "cdbfile:data/lists/$resource_name";
  $attr{match_type} = 'is';
  $attr{ui_editable} = 'yes';
  $resource_def{attributes} = \%attr;
  unless( PureMessage::Resources::Manager->DefineDB(%resource_def)) {
    $LOG->log('WARN',"Unable to defined a db resource with a name of $attr{name}");
    die "Can't Define"; 
  }
  unless( PureMessage::Resources::Manager->DefineLocal(%resource_def)) {
    $LOG->log('WARN',"Unable to defined a local resource with a name of $attr{name}");
    die "Can't Define"; 
  }
  my $store_db = PureMessage::Resources::Store::DB->new();
  my $store_lc = PureMessage::Resources::Store::Local->new();
  my $rm = PureMessage::Resources::Manager->new( store => $store_db );
  my $lm = PureMessage::Resources::Manager->new( store => $store_lc);
  $rm->add($resource_name);
  $lm->add($resource_name);
  $LOG->log('INFO',"Created resource definition and instance of '$attr{name}' with a description of '$attr{description}'");
  return $resource_name;

}

=item migrate_to_custom

 Migrates the group passed in from an Active Directory group to a local group

=cut

sub migrate_to_custom {
  my ($self, $resource_name) = @_;
  my $resource = $self->find($resource_name);
 
  my $list_id = PureMessage::Resources::Manager->CreateResourceTarget(
                            'list'          => 'custom',
                            'name'          => $resource->{attributes}{name},
                            'description'   => $resource->{attributes}{name}
                        ); 
  my $new_resource = $self->find($list_id);
  $new_resource->sync($resource, 1);
  my $store_db = PureMessage::Resources::Store::DB->new();
  my $store_lc = PureMessage::Resources::Store::Local->new();
  my $rm = PureMessage::Resources::Manager->new( store => $store_db );
  my $lm = PureMessage::Resources::Manager->new( store => $store_lc);
  $rm->sync($new_resource);
  $lm->sync($new_resource);

  $self->remove($resource_name);
  $self->DeleteResourceDefinition($resource_name);
  return $list_id;
}


1;


=back

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC

=head1 SEE ALSO
 
 none
  
=cut
