package PureMessage::Resources::Resource::Base;
use PureMessage::Resources::DBH;
use PureMessage::Resources::Location;
use PureMessage::Version qw($PREFIX);

use strict;
use warnings;
# common base package for all PM::R::R::imps

sub new {
    my ($CLASS,$store,%conf) = @_;
    my $self = bless(\%conf,$CLASS);
    $self->{store} = $store;
    $self->{filename} = "$PREFIX/etc/$self->{name}";
    $self->{pristine_filename} = 
        "$PREFIX/etc/data/resources/pristine/$self->{name}";
    $self;
}

# for bulk operations you should begin_batch at the start
# and end_batch to finish.  This should give a speed
# increase when manipulating many resources at one time

sub begin_batch {}
sub end_batch {}

# some resources need to have some work done to setup them up
# properly when local. do it here. by default we don't do anything
sub create_local { 1 }

sub name {
    my $self = shift;
    return $self->{resource_name} || $self->{name} || undef;
}

sub description {
    my $self = shift;
    return $self->{attributes}->{description} || $self->{name};
}

# this is just wrong, hname is the Human Name because the name as defined
# with resources is what we usually think of as the ID.  Name should be
# renamed appropriately so name can be what we expect it be.

sub hname {
    my $self = shift;
    return $self->{attributes}->{name} || $self->{name};
}

sub type {
    my $self = shift;
    return $self->{type} || undef;
}

my $SQL_RENAME_RESOURCE = <<EOF;
  UPDATE cfg_attribute
    SET attribute_value = ?
      WHERE resource_name = ?
        AND attribute_name like 'name';
EOF

my $SQL_INSERT_NAME_RESOURCE = <<EOF;
  INSERT INTO cfg_attribute(resource_name,attribute_name,attribute_value)
    VALUES(?,'name',?);
EOF

sub rename {
    my $self = shift;
    my $new_name = shift;
    unless($new_name) {
        die "Invalid name '$new_name' passed to resource rename.";
    }
    my $dbh = DBH();
    my $status = $dbh->do($SQL_RENAME_RESOURCE,undef,$new_name,$self->name);
    unless($status > 0) {
        # this resource doesn't have a name attribute
        # to update, so lets add one instead.
        $status = $dbh->do($SQL_INSERT_NAME_RESOURCE,
                           undef, $self->name, $new_name);
    }
    $self->{attributes}->{name} = $new_name;
    return $status;
}

sub native {
    my $self = shift;
    my %args = @_;
    
    my $writable = 1 if exists($args{writable}) and $args{writable};

    # if they desire a writable resource then don't give them a previously
    # cached one since we don't normally cache writable native resources
    
    if (!$writable and exists($self->{native}) and $self->{native} and $self->{native_pid} eq $$ ) {
        # don't make more than one of them and always make sure we refresh the
        # one we return unless it was a writable one
        $self->{native}->refresh() if $self->{native}->can('refresh') and ! $self->{store}->writable();
        return $self->{native};
    }
    
    # although if the entire resource manager is created writable then we
    # need to make writable native objects by default
    $args{writable} = 1 if !exists($args{writable}) and $self->{store}->writable();

    my $native;
    
    if($self->{store}->type eq 'local') {
        
        $native = $self->_native_local(%args);
        
    } elsif ($self->{store}->type eq 'db') {
        
        die "No instance_id defined for this resource, cannot created a native instance" unless $self->{instance_id};
        $native = $self->_native_db(%args);
        
    } else {
        
      die "Unknown store type '" . $self->{store}->type . "', resource '$self->{name}' can not be created";
      
    }
    
    # don't cache this if it is writable one time deal
    unless($writable) { 
        $self->{native} = $native;
        $self->{native_pid} = $$;
    }
    $native;
}

# the following methods should be overridden by implementation
# class

sub _native_local {}

sub _native_db {}

sub checksum {
    my $self = shift;
    die "Not Implemented";
}

sub change_id {
    my $self = shift;
    return $self->{store}->change_id($self->{name});
}

sub set_change_id {
    my ($self,$change_id) = @_;
    return $self->{store}->set_change_id($self->{name}, $change_id) 
        if $self->{store}->type eq 'local';
}

sub cclear {}
sub cinsert {}
sub cdelete {}
sub cupdate {}

sub enumerate {
    die "Enumerate not implemented";
}


sub clear_error {
    my $self = shift;
    delete $self->{_error};
}

sub error {
    my ($self,@args) = @_;
    if (@args == 1 ) {
        $self->{_error} = $args[0];
    } 
    else {
        return $self->{_error};
    }
}

sub sync {
    my ($self,$sync_from,$force_reinit) = @_;
    return unless $sync_from;
    return unless $self->type() eq $sync_from->type();
    
    my $count = 0;
    
    # skip sync if we have exactly the same version
    my $our_cid = $self->change_id;
    my $from_cid = $sync_from->change_id;
    if ($ENV{PMX_INCREMENTAL_DEBUG}) {
        print "change $our_cid to $from_cid\n";
        print "reinit forced\n" if $force_reinit;
    }
    my $need_update = 1;
    if ($from_cid and $from_cid == $our_cid) { 
        $need_update = 0;
        if ($sync_from->{store}->type eq 'local') {
            # check if the local files have changed
            $need_update = $sync_from->has_changed;
        }
    }
    return 0 unless $need_update or $force_reinit; 
    
    $self->autocommit(0); # batch transactions if possible
    my $is_incr;
    if ($force_reinit or !$self->can_incremental()) {
        $count = $self->sync_full_from($sync_from);
    } 
    else {
        $count = $self->sync_incremental_from($sync_from);  
        $is_incr++;
    }

    $self->save() if $self->can('save');
    $self->refresh() if $self->can('refresh');
    $self->autocommit(1);
    die "Error syncing " . $sync_from->name() . ": '" . $sync_from->error() . "'" if $sync_from->error();
    die "Error syncing " . $self->name() . ": '" . $self->error() . "'" if $self->error();
    if ($is_incr and $sync_from->{store}->type eq 'local') {
        # update the local change id
        $sync_from->set_change_id($self->change_id);
    }
    return $count;
}

my $SQL_UPDATE_LOCATION_CHANGE_ID = <<EOF;
  UPDATE cfg_location_instance
    SET last_change_id = ?
    WHERE location_name = ?
      AND instance_id = ?
EOF

sub change_id_to_location {
    my ($self,$location,$change_id) = @_;
    $change_id = 0 unless $change_id;
    return unless $self->{store}->type() eq 'db';
    return unless $location->location_name ne 'unknown' and $location->location_name();
    my $dbh = DBH();
    my $status = $dbh->do($SQL_UPDATE_LOCATION_CHANGE_ID, undef, $change_id,
                          $location->location_name(), $self->{instance_id});
    return $status;
}

sub sync_full_from {
    my ($self,$sync_from) = @_;
    my $count = 0;
    my $sub = sub {
        my @changes = @_;
        $self->cdo(@changes);
        $count++;
    };
    $sync_from->enumerate($sub);
    $self->cdo(['S', $sync_from->change_id]);
    $sync_from->change_id_to_location(PureMessage::Resources::Location->new(),
                                      $self->change_id());
    return $count;
}

sub can_incremental {
  return 0;
}

# ensure that the incremental diff data has all data from and including
# our version number, otherwise trigger a full resync
# 
my $SQL_VERIFY_CHANGE_TAIL = <<EOF;
 SELECT min(change_id) as change_tail 
   FROM cfg_item_change
   WHERE instance_id = ?
EOF

my $SQL_GET_DIFF = <<EOF;
 
 SELECT op, userkey, info1, info2, change_id 
   FROM cfg_item_change 
   WHERE instance_id = ?
     AND change_id > ?
   order by change_id asc 
EOF

sub change_diff_valid {
    my ($self,$instance_id,$change_id) = @_;
    my $dbh = DBH();
    my $sth = $dbh->prepare($SQL_VERIFY_CHANGE_TAIL);
    $sth->execute($instance_id);
    my $data = $sth->fetchrow_hashref();
    return unless $data->{change_tail};
    return 1 if $data->{change_tail} <= $change_id;
    return;
}

sub sync_incremental_from {
    my ($self,$sync_from) = @_;
    my $dest_change_id = $sync_from->change_id;
    print "incremental update started\n" if $ENV{PMX_INCREMENTAL_DEBUG};
    my $from_db = $sync_from->{store}->type eq 'db';
    my $can_incremental = 0;
    my $type = $sync_from->type;
    if ($from_db) {
        # make sure that we are able to update, ie. all changes from our
        # current version are there
        $can_incremental = $self->change_diff_valid($sync_from->{instance_id},
                                                    $self->change_id());
    }
    else {
        # check if this type of resource can be diffed using the pristine copy
        # We don't want to do this for other types right now... maybe later
        $can_incremental++ if $type eq 'list' or $type eq 'map' 
                              or $type eq 'multilist';
    }
    unless($dest_change_id and 
           $sync_from->can_incremental() and
           $can_incremental) {
        print "could not do an incremental update, trying a full.\n" 
            if $ENV{PMX_INCREMENTAL_DEBUG};
        return $self->sync_full_from($sync_from);
    }

    my $start_version = $self->change_id() || 0;
    my $instance_id = $sync_from->{instance_id};

    if ($from_db) {
        my $dbh = DBH();
        my $sth = $dbh->prepare($SQL_GET_DIFF);
        $sth->execute($instance_id, $start_version);
        my $change_id;
        my $count = 0;
        while(my $data = $sth->fetchrow_arrayref) {
            $self->cdo([@$data]);
            $count++;
            $change_id = $data->[4];
            $self->cdo(['S',$change_id]) if $count % 10 == 0;
        }
        $change_id = $dest_change_id unless $change_id; 
        # if we have no rows but are at a particular version 
        # we need to increment the version to match
        $self->cdo(['S', $change_id]);
        $sync_from->change_id_to_location(PureMessage::Resources::Location->new(),
                                          $self->change_id());
        return $count;
    }
    else { # sync from local to db ($self is db)
        my $changes = $sync_from->local_changes;
        my $count = 0;
        for my $c (@{$changes->{'-'}}) {
            $count++;
            $self->cdo(['D', @$c]);
        }
        for my $c (@{$changes->{'+'}}) {
            $count++;
            $self->cdo(['I', @$c]);
        }
        return $count;
    }
}

sub cfmt_clear {
    my $self = shift;
    return ['C','','',''];
}

sub cfmt_delete {
    my ($self,$user,$key,$value) = @_;
    return ['D',$user,$key,$value];
}

sub cfmt_clone {
    my ($self,$user,$key,$obj) = @_;
    return ['L',$user,$key,$obj];
}


sub cfmt_insert {
    my ($self,$user,$key,$value) = @_;
    return ['I',$user,$key,$value];
}

# wether or not this resource should batch until a save or write
# as soon as it possibly can
sub autocommit {
    my ($self,$auto) = @_;
    $self->{autocommit} = $auto;
}

sub save {
    my $self = shift;
}

sub cdo {
    my ($self,@records) = @_;
    foreach my $r (@records) {
        print "change(" . join(',', map { defined $_ ? $_ : '' } @$r) 
                        . ")\n" if $ENV{PMX_INCREMENTAL_DEBUG};
        my $op = shift(@$r);
        if ($op eq 'I') {
            $self->cinsert(@$r);
        } elsif ($op eq 'C') {
            $self->cclear(@$r);
        } elsif ($op eq 'D') {
            $self->cdelete(@$r);
        } elsif ($op eq 'U') {
            $self->cupdate(@$r);
        } elsif ($op eq 'S') {
            $self->set_change_id(@$r);
        } elsif ($op eq 'L') {
            $self->cclone(@$r);
        } else {
            die "Unknown diff record '$op'";
        }
    }
}

# this is a helper sub that is passed into the enumerate of to_text
# it should just return an array of lines that represent the passed
# data.  If the data shouldn't be show, this sub should return undef.

sub _aggregate_op {
    my ($self,$op,$a,$b,$c) = @_;
    # this will likely only work for lists
    return($b) if $op eq 'I'; 
    return; 
}

# return the contents of this resource as a text stream formatted for easy
# display to endusers
sub to_text {
    my $self = shift;
    my @OUTPUT = ();

    $self->enumerate(sub { my @agg_output = $self->_aggregate_op(@{$_[0]}); 
                           push(@OUTPUT,@agg_output) if scalar(@agg_output)
                     });
    
    push(@OUTPUT,$self->error()) if $self->error();
    return @OUTPUT;
}


# Used for updating from local resource mgr to db
# subclasses can do local file checks with the pristine copy
sub has_changed { 0 }

# Used for updating from local resource to db
# The pristine area keeps copies of the resources which
# are the same as in the db, so that we can make incremental
# changes to the db
sub update_pristine {}

# Used for updating from local resource to db
# This function should examine the local files
# and return an hashref containing the changes.  There may
# be two keys: '-' and '+'.  The corresponding values should
# be an arrayref containing an arrayref suitable for passing to cdo
sub local_changes { die "Not implemented" }

sub _diff_to_changes {
    my ($self, $diff, $userkey) = @_;
    # skip the diff headers, hunk headers and non-changed lines
    my @diff = grep { !/^(\-\-\-|\+\+\+|\@\@|\s)/ }
               split "\n", $diff;

    my $changes = {};
    for my $line (@diff) {
        unless ($line =~ /^([-+])(.*)/) {
            warn "Can't read line ($line)";
            next;
        }
        my ($change, $item) = ($1, $2);
        next if $item =~ /^\s*$/;
        push @{$changes->{$change}}, [$userkey, $item];
    }

    return $changes;
}

1;
