#!perl

package PureMessage::Resources::Resource::blob;

use strict;
use warnings;
use base qw(PureMessage::Resources::Resource::Base);

use Carp qw(croak);
use ActiveState::File::Atomic;
use DBD::Pg;

use PureMessage::Version qw($PREFIX);
use PureMessage::Resources::DBH qw(DBH);

# Queries

my $SQL_GET_BINVAL = <<"EOT";
    SELECT b_value FROM cfg_binval WHERE instance_id = ?
EOT

my $SQL_NUKE_BINVAL = <<"EOT";
    DELETE FROM cfg_binval WHERE instance_id = ?
EOT

my $SQL_SET_BINVAL = <<"EOT";
    INSERT INTO cfg_binval (instance_id, b_value) VALUES (?, ?)
EOT

return 1;

# Resource-y methods

sub can_incremental {
    return 0;
}

sub cclear {
    my $self = shift;
    my $dbh = $self->_dbh;
    $self->native->commit_string('');
}

sub cdelete {
    cclear(@_);
}

sub cinsert {
    my $self = shift;
    $self->native->commit_string($_[2]);
    if($self->{store}->type eq 'local' &&
        $self->{attributes}->{no_empty_file} &&
        !-s($self->_filename)
    ) {
        unlink($self->_filename);
    }
    delete($self->{native});
}

sub cupdate {
    cinsert(@_);
}

sub enumerate {
    my ($self, $sub, $change_id) = @_;
    $sub->($self->cfmt_insert('', '', $self->native->slurp));
}

# methods that emulate ActiveState::File::Atomic's behaviour
# (only as much as we need to get away with this method of syncing)


sub slurp {
    my $self = shift;
    return (
        $self->_dbh->selectrow_array($SQL_GET_BINVAL, {}, $self->{instance_id})
    )[0];
}

sub commit_string {
    my $self = shift;

    croak ref($self), "->commit_string() called with 'writable' turned off"
        unless($self->{native_args}->{writable});

    my $dbh = $self->_dbh;

    $dbh->begin_work if($dbh->{AutoCommit});
    $dbh->do($SQL_NUKE_BINVAL, {}, $self->{instance_id});
    my $sth = $dbh->prepare($SQL_SET_BINVAL);
    $sth->bind_param(1, "$self->{instance_id}");
    $sth->bind_param(2, $_[0], { pg_type => DBD::Pg::PG_BYTEA() });
    $sth->execute();
    $dbh->commit;
}

# methods that only our base class should call

sub _native_local {
    my $self = shift;
    my %args = @_;
    my $filename = $self->_filename;
    return ActiveState::File::Atomic->new(
        $filename, writable => $args{writable} || 0,
        create => $args{writable} || 0,
    );
}

sub _native_db {
    my $self = shift;
    my %args = @_;
    $self->{native_args} = \%args;
    $self->_dbh;
    $self;
}

# internal methods

sub _dbh {
    my $self = shift;
    return $self->{dbh} if($self->{dbh} && $self->{dbh}->ping);
    $self->{dbh} = DBH() || die "Unable to initialize database connection";
    $self->{dbh};
}

sub _filename {
    my $self = shift;
    my $filename = $self->{attributes}->{filename} =~ /^[^\/]/ 
                 ? "$PREFIX/$self->{attributes}->{filename}" 
                 : $self->{attributes}->{filename};
    $filename;
}

sub _aggregate_op {
  my ($self,$op,$a,$b,$c) = @_;
  # blobs are binary
  return;
}

# from the department of redundancy department

sub DESTROY {
    my $self = shift;
    if(defined($self->{native_args})) {
        my $x = $self->{native_args};
        delete($self->{native_args});
    }
}
