package PureMessage::Resources::Store::DB;
use PureMessage::Resources::Config qw(%config);
use PureMessage::Resources::DBH qw(DBH);
use PureMessage::Resources::Resource;
use PureMessage::Resources::Manager;
use PureMessage::Resources::Location;
use DBI;
use base PureMessage::Resources::Store::Base;

sub new {
    my ($CLASS, %options) = @_;

    my $self = {};
    bless($self,$CLASS);
    # store::DB requires that this location be registered before any operations
    # on db hosted data can be undertaken.
    my $location = PureMessage::Resources::Location->new();
    die "You must register this location before you can use a store of type db\n"
      unless $location->location_name and $location->location_name ne 'unknown';
    $self->{dbh} = DBH();   
    $self->{writable} = $options{writable} if exists($options{writable});
    return $self;
}
sub begin_batch {
}

sub end_batch {
}

sub type {
    return 'db';
}

my $SQL_RESOURCE_ENUMERATE = <<EOF;
  SELECT cfg_instance.resource_name as name, instance_id, resource_type as type
        FROM cfg_instance, cfg_resource as res
        WHERE cfg_instance.resource_name = res.resource_name
EOF
# XXX TODO: restrict the above to not syncronize resources based on the
# existance of a flag in the resource's attributes

my $DEFAULT_PROFILE_ID_FOR_PROTOTYPE = 0;

sub enumerate {
    my ($self,$sub) = @_;
    return unless ref($sub) eq 'CODE';
    my $dbh = $self->_dbh();
    my $sth = $dbh->prepare($SQL_RESOURCE_ENUMERATE);
    $sth->execute;

    while($data = $sth->fetchrow_hashref) {
        my $obj = PureMessage::Resources::Resource->new($self,instance_id => $data->{instance_id},%$data);
        &$sub($obj);
    }

}

# initialize our connection to the database if it's not opened. Otherwise
# return the cached value we have.
sub _dbh {
   my $self = shift;
   return $self->{_dbh} if $self->{_dbh} and $self->{_dbh}->ping;
   $self->{_dbh} = DBH();
   die "Unable to initialize database connection" unless $self->{_dbh};
   return $self->{_dbh};    
}


my $SQL_CREATE_RESOURCE_INSTANCE = <<EOF;
  insert into cfg_instance (resource_name,profile_id)
       values (?,0)

EOF

my $SQL_REMOVE_RESOURCE_INSTANCE = <<EOF;
  DELETE FROM cfg_instance 
       WHERE resource_name = ?
EOF

my $SQL_FIND_RESOURCE_INSTANCE = <<EOF;
  select instance_id, inst.resource_name, profile_id, resource_type as type
        from cfg_instance as inst, cfg_resource as res
        where inst.resource_name = ?
          and res.resource_name = inst.resource_name
EOF

my $SQL_GET_ATTRIBUTES = <<EOF;
  select attribute_name, attribute_value 
        from cfg_attribute
        where cfg_attribute.resource_name = ?
EOF

sub get_attributes {
  my ($self,$rname) = @_;
  my $dbh = $self->_dbh();
  my $sth = $dbh->prepare($SQL_GET_ATTRIBUTES);
  my $found = 0;
  $sth->execute($rname);
  my %attr;
  while (my $r = $sth->fetchrow_hashref) {
      $attr{$r->{attribute_name}} = $r->{attribute_value};
      $found++;
  }
  return \%attr if $found;
  return;
}

sub find {
    my ($self,$name,%options) = @_;
    my $cache = $self->SUPER::find($name);
    return $cache if $cache;
    # attempt a db connection only when it's required.    
    my $dbh = $self->_dbh();
    my $sth = $dbh->prepare($SQL_FIND_RESOURCE_INSTANCE);
    $sth->execute($name);
    my $data = $sth->fetchrow_hashref();
    return unless $data;
    $data->{name} = $data->{resource_name};
    $data->{type} = $options{type} if $options{type};
    $data->{ignore_errors} = 1 if $options{ignore_errors};
    my $attr = $self->get_attributes($name);
    $data->{attributes} = $attr if $attr;
    my $res =  PureMessage::Resources::Resource->new($self,%$data); 
    $self->found($res);
    return $res;
}

sub add {
    my ($self,$name,%options) = @_;
    # attempt a db connection only when it's required.    
    my $data = $self->find($name);
    unless($data) {
        my $def = PureMessage::Resources::Manager->Find($name);
        die "Undefined Resource '$name'\n" unless $def;
        my $dbh = $self->_dbh();
        my $sth = $dbh->prepare($SQL_CREATE_RESOURCE_INSTANCE);
        $sth->execute($name);
        $data = $self->find($name,%options); 
    }
    $self->found($data);
    return $data; # return a resource object
}

sub remove {
    my ($self,$name) = @_;
    return unless $name;
    my $dbh = $self->_dbh();
    $self->forget($name);
    return $dbh->do($SQL_REMOVE_RESOURCE_INSTANCE,undef,$name);
}

sub set_change_id {
    my ($self,$resource_name,$change_id)  = @_;
    # this is an error attempting to set the change_id on a db resource
    die "DB hosted resources cannot be forced to a specific change_id\n";
}

my $SQL_GET_CHANGE_ID = <<EOF;
 SELECT change_id FROM cfg_instance 
   WHERE resource_name = ? 
EOF

sub change_id {
    my ($self,$resource_name) = @_;
    my $dbh = $self->_dbh();
    my $sth = $dbh->prepare($SQL_GET_CHANGE_ID);
    $sth->execute($resource_name);

    my $data = $sth->fetchrow_hashref;
    return $data->{change_id} if $data->{change_id};
    return;
}


sub sync {
    my $self = shift;
    my $sync_from = shift;
    my %opts = @_;
    return unless $sync_from;
    my $target = $self->find($sync_from->name(),%opts);
    return unless $target; # TODO add it instead????
    return unless $target->type() eq $sync_from->type();
    return $target->sync($sync_from,$opts{force} ? 1 : 0);

}

1;
