package PureMessage::Resources::Store::Local;
use PureMessage::Resources::Config qw(%config);
use PureMessage::Resources::KVData::Local;
use ActiveState::Config qw(:path);
use PureMessage::Resources::Resource;
use PureMessage::Resources::Manager;
use ActiveState::Config::Simple qw(config_parse_file);
use Data::Dumper;
use base PureMessage::Resources::Store::Base;
use strict;


my $CONFIG_FILE = 'instances.conf';

sub new {
    my ($CLASS, %options) = @_;
    my $self = { _instances => undef };
    bless($self,$CLASS);
    $self->{writable} = $options{writable} if exists($options{writable});
    return $self;
}

sub type {
    return 'local';
}

sub begin_batch {
  my $self = shift;
  $self->{batch} = undef;
  $self->_init_as(writable => 1);
  $self->{batch} = 1;
}

sub end_batch {
   my $self = shift;
   delete($self->{batch});
   $self->{as}->save() if $self->{as};
   delete($self->{as});
}

sub _is_batch {
   my $self = shift;
   return 1 if $self->{batch};
   return;
}

sub _init_as {
    my $self = shift;
    my %options = @_;
    die "Invalid batch operation, Config not initialized" if $self->{batch} and not $self->{as};
    return if $self->{batch};
    # have to destroy the old AS::Config and free it's lock before making a new
    # one for the same file
    delete($self->{as});
    my $as = ActiveState::Config->new(strict => 0, allow_includes => 1);
    my $filename = "$config{prefix}/etc/$CONFIG_FILE";
    $as->parse_file($filename,%options);
    $self->{as} = $as;
    $self->{_instances} = undef;
}

sub _as_hash {
    my $self = shift;
    return $self->{_instances} if $self->{_instances};
    if ($self->{as}) {
      $self->{_instances} = $self->{as}->as_hash();
    } else {
      $self->{_instances} = config_parse_file("$config{prefix}/etc/$CONFIG_FILE");

    }
    return $self->{_instances};
}


sub _add_instance {
    my ($self,$name) = @_;
    my $data = $self->find($name);
    unless($data) {
        my $def = PureMessage::Resources::Manager->FindLocal($name);
        die "Undefined Resource '$name'\n" unless $def;
        my $temp_resource = PureMessage::Resources::Resource->new($self,%$def);
        die "Unable to create local storage for '$name'\n" 
                unless $temp_resource->create_local();
        $self->forget($name); # so we don't return this half broken one later
        $self->_init_as(writable => 1) unless $self->_is_batch();
        my $as = $self->{as};
        my $root = $as->root();
        path_value($root,"instance#$def->{name}.type",$def->{type});
        path_value($root,"instance#$def->{name}.name",$def->{name});
        $self->{_instances}->{instance}->{$def->{name}}->{name} = $def->{name};
        $self->{_instances}->{instance}->{$def->{name}}->{type} = $def->{type};
        # don't save extra data
        #foreach my $k (keys %{$def->{attributes}}) {
        #  path_value($root,"instance#$def->{name}.attributes.$k",$def->{attributes}->{$k});
        #}
        unless( $self->_is_batch() ) {
            $as->save();
            $self->{as} = undef;
            $self->_init_as();
            $self->{_instances} = undef;
        }
        $data = $self->find($name); 
    }
    return $data; # return a resource object
}

sub _find_instance {
    my ($self,$name) = @_;
    my $conf = $self->_as_hash();
    return unless $conf and ref($conf) eq 'HASH';
    return unless defined $conf->{instance}->{$name}; # we have a resource, get it's data
    # get all the info about this resource from our local config file
    my $def = $config{resource}->{$name};
    die "Can't find a resource definition for a local resource ($name)\n" unless $def;
    $def->{change_id} = exists($conf->{change_id}) ? $conf->{change_id} : undef;
    return $def;
}

sub enumerate {
    my ($self,$sub) = @_;
    my $conf = $self->_as_hash();
    return unless ref($sub) eq 'CODE';
    foreach my $res (keys %{ $conf->{instance} } ) {
        my $obj = $self->find($res);
        return unless $obj;
        &$sub($obj);
    }
}

sub find {
    my ($self,$name,%options) = @_;
    my $cache = $self->SUPER::find($name);
    return $cache if $cache;
    my $data = $self->_find_instance($name);
    return unless $data;
    $data->{ignore_errors} = 1 if $options{ignore_errors};
    my $res = PureMessage::Resources::Resource->new($self,%$data); 
    $self->found($res);
}

sub add {
    my ($self,$name) = @_;
    my $data = $self->_add_instance($name);
    return unless $data;
    $self->found($data);
    return $data;
}

sub remove {
    my ($self,$name) = @_;
    $self->_init_as(writable => 1);
    my $as = $self->{as};
    my $root = $as->root();
    $self->forget($name);
    path_node($root,"instance#$name",delete => 1);
    $as->save();
    $self->_init_as();
}

sub set_change_id {
    my ($self,$resource_name,$change_id)  = @_;
    return unless defined($change_id);
    $self->_init_as(writable => 1) unless $self->_is_batch();
    my $as = $self->{as};
    my $root = $as->root();
    path_value($root,"instance#$resource_name.change_id",$change_id);
    $self->{_instances}->{instance}->{name}->{change_id} = $change_id;
    unless($self->_is_batch()) {
      $as->save();
      $self->{as} = undef;
    }

}

sub change_id {
    my ($self,$resource_name)  = @_;

    unless($self->{as}) {
      $self->_init_as(writable => 1);
    }

    my $as = $self->{as};
    my $root = $as->root();
    my $change_id = path_value($root,"instance#$resource_name.change_id");
    return $change_id || 0;
}

sub sync {
    my $self = shift;
    my $sync_from = shift;
    my %opts = @_;
    return unless $sync_from;
    my $target = $self->find($sync_from->name(),%opts);
    unless($target) {
        # we don't have that item we need to create it, we depend
        # on there being a resource definition for this resource
        # then add will create it for us.
        my $res = $self->add($sync_from->name());
        die "Unable to create a local instance of '" . $sync_from->name() . "'\n" unless $res; 
    }; 
    return unless $target->type() eq $sync_from->type();
    return $target->sync($sync_from,$opts{force} ? 1 : 0);
    
}




1;


